/*
        PKTMAKE.C       - FTN type 2+ packets maker ver. 1.7

        Copyright (C) 1994 D.Marck, FIDONet 2:5020/268

        Last revised: 04-Jun-95

        Functions:
        wrtbuf          pattr           makepkt         getaddr
        getattr         doopt           (strichr)       getrespfile
        main

        29-May-94       . Getting started
                        * Ver. 1.0
                        - PKT Capability word bug fixed
        30-May-94       * Ver. 1.1
                        - Path/SEEN-BY bug fixed
                        * Ver. 1.2
                        - Fixed VERY stupid bug with SEEN-BY
        04-Sep-94       * Ver. 1.3
                        - Small path bug fixed
                        * Ver. 1.4
                        - Fixed stupid bug with SEEN-BYs
        08-Sep-94       * Ver. 1.5
                        - Finally fixed bug with Type 2+ CW
        29-Sep-94       * Ver. 1.6
                        - PATH & SEEN-BYs changed
                        - CR/LF added rather than just CR
                        - Curious bug with FMPT/TOPT/INTL fixed
                        - Bug with MSGID (always .0) fixed
                        . ^aPID format slightly fixed
        04-Jun-95       * Ver. 1.7
                        - Origin zero-point bug fixed
*/

#include        <stdio.h>
#include        <string.h>
#include        <ctype.h>
#include        <io.h>
#include        <dir.h>
#include        <fcntl.h>
#include        <sys\stat.h>
#include        <time.h>

#include        "ftn.h"
#include        "getopt.h"


#define         VERSION         "1.7"
#define         VERSHI          1
#define         VERSLO          6
#define         PGMNAME         "PktMake"
#define         VERSIONSTR      PGMNAME " ver. " VERSION
#define         PIDSTR          PGMNAME " " VERSION

#define         DEFTO           "All"
#define         DEFFROM         PGMNAME
#define         DEFSUBJ         ""
#define         DEFORIGIN       " ***** Automatically posted message ***** "


char    title[] = "PktMake - FTN type 2+ packets maker ver. " VERSION ", "
                  __DATE__ " " __TIME__ "\n"
                  "Copyright (C) 1994 D.Marck, FIDONet 2:5020/268\n\n";

char    usage[] = "Usage: PktMake [params] [@respfile] [path to PKT]"
                        " [< file to post]\n"
                  "Parameters (all may be used in response file):\n"
                  "     -D<addr>        packet destination address\n"
                  "             should be used first, which allows other"
                        " addresses to be shortened\n"
                  "     -P<addr>        packet originating address\n"
                  "     -O<addr>        message originating address, "
                        "defaults to packet orig\n"
                  "     -M<addr>        message destination address, "
                        "defaults to packet dest\n"
                  "     -T<name>        to field, defaults to '"
                        DEFTO "'\n"
                  "     -F<name>        from field, defaults to '"
                        DEFFROM "'\n"
                  "     -S<name>        subject, defaults to '"
                        DEFSUBJ "'\n"
                  "     -A<tag>         areatag, defaults to none\n"
                  "     -R<origin>      origin, defaults to\n"
                  "              '" DEFORIGIN "'\n"
                  "     -W<password>    packet password\n"
                  "     -B<attrchars>   message attributes (PCFRA: "
                                "Private, Crash, File, RRQ, ARQ)\n"
                  "             defaults to P in netmail, none in echo\n"
                  "     -I              force INTL kludge generation\n"
                  "     -X              do not use tearline\n";


NETADDR dstaddr,                /* pkt destination address */
        pktaddr,                /* pkt origination address */
        msgaddr,                /* msg origination address */
        mtoaddr;                /* msg destination address */

char    to[MAXTOLEN]         = DEFTO,
        from[MAXFROMLEN]     = DEFFROM,
        subj[MAXSUBJLEN]     = DEFSUBJ,
        origin[MAXORIGINLEN] = DEFORIGIN,
        area[80]             = "";

char    password[MAXPWD] = "";

char    pktpath[MAXPATH] = "";

BYTE    forcintl = NO,
        tearline = YES;
WORD    msgattr  = 0;

PKTHDR  phdr;
PMSGHDR mhdr;

static  char    buff[4096];


void    wrtbuf( int fd, void *buf, WORD len )
{
        if ( write( fd, buf, len ) != len )
            error( 254, "Write error. Disk full?" );
}

void    pattr( WORD attr, WORD mask, char *s )
{
        if ( attr & mask )
            fputs( s, stderr );
}

void    makepkt( void )
{
        time_t  tm;
        struct  tm      *ltm;
        char    *s;
        int     fd, len;
        WORD    origp, destp;


        tm = time( NULL );

        if ( pktpath[0] && LASTCHAR(pktpath) != ':' )
            strcat( pktpath, "\\" );

        s = &LASTCHAR(pktpath)+1;
        do {
            sprintf( s, "%08lX.PKT", tm );

            fd = open( pktpath, O_RDONLY|O_BINARY|O_DENYNONE );
            if ( fd != ERROR )
            {
                warning( "%s exists; trying next name", pktpath );
                tm++;
                close( fd );
            }
        } while ( fd != ERROR );
        if ( (fd = open( pktpath, O_WRONLY|O_BINARY|O_CREAT|O_TRUNC|O_EXCL|
                O_DENYWRITE, S_IREAD|S_IWRITE )) == ERROR )
                error( 254, "Can't create packet %s", pktpath );

        /* form packet header */
        phdr.pkttype = PKTTYPE;
        ltm = localtime( &tm );

        phdr.year   = ltm->tm_year + 1900;
        phdr.month  = ltm->tm_mon;
        phdr.day    = ltm->tm_mday;
        phdr.hour   = ltm->tm_hour;
        phdr.min    = ltm->tm_min;
        phdr.sec    = ltm->tm_sec;
        phdr.baud   = 0;

        phdr.qm_origzone =
        phdr.origzone   = pktaddr.zone;
        phdr.qm_destzone =
        phdr.destzone   = dstaddr.zone;
        phdr.auxnet     =
        phdr.orignet    = pktaddr.net;
        phdr.destnet    = dstaddr.net;
        phdr.orignode   = pktaddr.node;
        phdr.destnode   = dstaddr.node;
        phdr.origpoint  = pktaddr.point;
        phdr.destpoint  = dstaddr.point;

        phdr.prodcode     = DEFPRODCODE;
        phdr.prodcodehigh = 0;
        phdr.rev_lev      = VERSHI;
        phdr.revminor     = VERSLO;
        memset( phdr.password, '\0', MAXPWD );
        strncpy( phdr.password, password, MAXPWD );
        strncpy( (char *)&phdr.pr_data, DEFPRDATA, sizeof(DWORD) );

        phdr.CW      = CW_TYPE2;
        phdr.CW_copy = CW_TYPE2COPY;


        /* form message header */

        if ( !msgaddr.net )
            msgaddr = pktaddr;
        if ( !mtoaddr.net )
            mtoaddr = dstaddr;

        mhdr.pkttype  = PKTTYPE;
        mhdr.orignet  = msgaddr.net;
        mhdr.destnet  = mtoaddr.net;
        mhdr.orignode = msgaddr.node;
        mhdr.destnode = mtoaddr.node;

        mhdr.attr = msgattr;
        mhdr.cost = 0;
        pmsgdate( mhdr.datetime, ltm );

        /* make variable length header part & kludges */
        s = buff;
        s = stpcpy( s, to ) + 1;
        s = stpcpy( s, from ) + 1;
        s = stpcpy( s, subj ) + 1;
        if ( area[0] )
            s += sprintf( s, "%s%s\r", KLDG_AREA, area );

        origp = msgaddr.point;
        destp = mtoaddr.point;

        if ( forcintl || msgaddr.zone != mtoaddr.zone )
        {
            mtoaddr.point = msgaddr.point = 0;
            s += sprintf( s, "%s%s", KLDG_INTL, paddr(&mtoaddr) );
            s += sprintf( s, " %s\r", paddr(&msgaddr) );
        }
        if ( origp )
            s += sprintf( s, "%s%d\r", KLDG_FMPT, origp );
        if ( destp )
            s += sprintf( s, "%s%d\r", KLDG_TOPT, destp );
        msgaddr.point = origp;
        s += sprintf( s, "%s%s %08lX\r", KLDG_MSGID, paddr(&msgaddr), tm );
        msgaddr.point = 0;
        if ( !tearline )
            s += sprintf( s, "%s%s\r", KLDG_PID, PIDSTR );

        /* write it all down */
        wrtbuf( fd, &phdr, PKTHDRSZ );
        wrtbuf( fd, &mhdr, PMSGHDRSZ );
        wrtbuf( fd, &buff, s - buff );

        fprintf( stderr, "Writing %s: %s -> ", pktpath, paddr( &pktaddr ) );
        fprintf( stderr, "%s\n", paddr( &dstaddr ) );
        fprintf( stderr, "From %s (%s) ", from, paddr( &msgaddr ) );
        fprintf( stderr, "to %s (%s)\n", to, paddr( &mtoaddr ) );
        fprintf( stderr, "Subj: %s\n", subj );
        if ( msgattr )
        {
            fprintf( stderr, "Attr: " );
            pattr( msgattr, MSGPRIVATE, "Pvt " );
            pattr( msgattr, MSGCRASH,   "Cra " );
            pattr( msgattr, MSGFILE,    "Att " );
            pattr( msgattr, MSGRRQ,     "RRQ " );
            pattr( msgattr, MSGARQ,     "ARQ " );
            fprintf( stderr, "\n" );
        }
        if ( area[0] )
            fprintf( stderr, "Area: %s\n", area );

        /* write stdin */
        while ( gets( buff ) )
        {
            len = strlen( buff );
            strcpy( &buff[len++], "\r" );
            wrtbuf( fd, buff, len );
        }

        /* tearline */
        if ( tearline )
            len = sprintf( buff, "%s %s\r", TEARLINE, VERSIONSTR ),
            wrtbuf( fd, buff, len );
        else if ( area[0] )
            len = sprintf( buff, "%s\r", TEARLINE ),
            wrtbuf( fd, buff, len );

        /* origin, path & seen-by */
        if ( area[0] )
        {
            int l;

            len = sprintf( buff, "%s%s", ORIGIN, origin );
            msgaddr.point = origp;
            l = strlen( s = paddr( &msgaddr ) );
            msgaddr.point = 0;
            if ( len + l + 3 > MAXORIGINLEN )
                len = MAXORIGINLEN - l - 3;
            len += sprintf( &buff[len], " (%s)\r\n", s );
            s = &buff[len];

            msgaddr = pktaddr;
            mtoaddr = dstaddr;

            /* Path/SEEN-BYs is 2D */
            msgaddr.zone  = mtoaddr.zone =
            msgaddr.point = mtoaddr.point = 0;

            if ( msgaddr.net  != mtoaddr.net ||
                 msgaddr.node != mtoaddr.node )         /* write path */
            {
                NETADDR *lo, *hi;

                if ( msgaddr.net < mtoaddr.net )
                    lo = &msgaddr, hi = &mtoaddr;
                else if ( msgaddr.net > mtoaddr.net )
                    lo = &mtoaddr, hi = &msgaddr;
                else if ( msgaddr.node < mtoaddr.node )
                    lo = &msgaddr, hi = &mtoaddr;
                else
                    lo = &mtoaddr, hi = &msgaddr;

                s += sprintf( s, "%s%s", KLDG_SEENBY, paddr( lo ) );
                s += sprintf( s, " %s\r\n", pshortaddr( hi, lo ) );
                s += sprintf( s, "%s%s\r\n", KLDG_PATH, paddr( &msgaddr ) );
            } else {
                s += sprintf( s, "%s%s\r\n", KLDG_SEENBY, paddr( &msgaddr ) );
            }

            wrtbuf( fd, buff, s - buff );
        }

        tm = 0;
        wrtbuf( fd, &tm , sizeof(BYTE) + sizeof(WORD) );

        close( fd );
}

void    getaddr( char *param, NETADDR *addr )
{
static  BYTE wasdst = 0;
        int     shortok;
        NETADDR *defaddr;


        defaddr = (addr == &dstaddr) ? NULL : &dstaddr;
        shortok = (addr == &dstaddr) ? -1   : 1;

        if ( !wasdst && shortok == 1 )
            error( 254, "Destination address should be the first" );
        if ( parseaddr( addr, param, shortok, defaddr ) == NULL )
            error( 254, "Bad address: %s", param );
        if ( addr == &dstaddr )
            wasdst = 1;
}

void    getattr( char *s )
{
        char    c;
        WORD    attr;

        for ( attr = 0; (c = *s) != '\0'; s++ )
        {
            switch ( toupper(c) )
            {
            case 'P':   attr |= MSGPRIVATE;     break;
            case 'C':   attr |= MSGCRASH;       break;
            case 'F':   attr |= MSGFILE;        break;
            case 'R':   attr |= MSGRRQ;         break;
            case 'A':   attr |= MSGARQ;         break;
            default:    error( 254, "Bad message attr: %c", c );
            }
        }
        msgattr = attr;
}

void    doopt( char opt, char *param )
{
        switch ( opt )
        {
        case 'D':       getaddr( param, &dstaddr );     break;
        case 'P':       getaddr( param, &pktaddr );     break;
        case 'O':       getaddr( param, &msgaddr );     break;
        case 'M':       getaddr( param, &mtoaddr );     break;
        case 'T':       ncpy( to,   param, MAXTOLEN );      break;
        case 'F':       ncpy( from, param, MAXFROMLEN );    break;
        case 'S':       ncpy( subj, param, MAXSUBJLEN );    break;
        case 'A':       strcpy( area, param );
                        strupr( area );
                        break;
        case 'R':       ncpy( origin, param, MAXORIGINLEN ); break;
        case 'W':       strncpy( password, param, MAXPWD ); break;
        case 'I':       forcintl = YES;                     break;
        case 'B':       getattr( param );                   break;
        case 'X':       tearline = NO;                      break;
        default:        error( 254, "Bad option: %c", opt );
        }
}

static  char    *strichr( const char *str, int c )
{
        c = toupper( c );
        while ( *str )
            if ( toupper( *str++ ) == c )
                return (char *)str;
        return NULL;
}

int     getrespfile( char *fn, char *paropt, void (*do_opt)(char, char *) )
{
        FILE    *fp;
static  char    buf[512];
        char    c, *delim, *s;

        char    *(*findfunc)( const char *, int );

        findfunc = ( getopt_case ) ? strchr : strichr;

        if ( (fp = fopen( fn, "r" )) == NULL )
            return ERROR;

        while ( fgets( buf, 512, fp ) )
        {
            for ( s = skip( buf ); *s; s = skip( s ) )
            {
                delim = NULL;
                if ( *s == '"' )
                    delim = "\"", s++;
                if ( *s == '-' || *s == '/' )
                {
                    for ( s++; (c = *s) != '\0'; s++ )
                    {
                        if ( !getopt_case )
                            c = toupper( c );

                        if ( delim && strchr( delim, c ) )
                            continue;
                        if ( !delim && isspace( c ) )
                            continue;

                        if ( findfunc( paropt, c ) )
                        {
                            char *t = s + 1;

                            while ( *t && (delim && !strchr( delim, *t ) ||
                                    !delim && !isspace(*t)) )
                                t++;
                            if ( !*t )
                                t[1] = '\0';
                            *t = '\0';
                            do_opt( c, s + 1 );
                            s = t + 1;
                            break;
                        }
                        else
                            do_opt( c, NULL );
                    }
                } else
                    error( 254, "option char expected" );
            }
        }

        fclose( fp );
        return OK;
}



static  char    longopt[] = "DPOMTFSARWB";

long    *nullptr = NULL;

void    main( int argc, char *argv[] )
{

        fputs( title, stderr );

        if ( argc == 1 )
           printf( usage ),
           error( 254, "" );

        argc = getopt( argc, argv, longopt, doopt );

        if ( argc > 1 && *argv[1] == '@' )      /* responsefile */
        {
            if ( getrespfile( &argv[1][1], longopt, doopt ) == ERROR )
                error( 254, "Can't open %s", &argv[1][1] );
            argc--, argv++;
        }

        if ( !dstaddr.net || !pktaddr.net )
            error( 254, "Address(es) missing" );

        if ( argc > 1 )
            zaptailbs( strcpy( pktpath, argv[1] ) );

        if ( !area[0] && !msgattr )
            msgattr = MSGPRIVATE;

        makepkt();
}

/*      End of PKTMAKE.C      */
